# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import time
from collections.abc import Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal

from airflow.configuration import conf
from airflow.providers.common.compat.sdk import BaseOperator
from airflow.providers.openai.exceptions import OpenAIBatchJobException
from airflow.providers.openai.hooks.openai import OpenAIHook
from airflow.providers.openai.triggers.openai import OpenAIBatchTrigger

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class OpenAIEmbeddingOperator(BaseOperator):
    """
    Operator that accepts input text to generate OpenAI embeddings using the specified model.

    :param conn_id: The OpenAI connection ID to use.
    :param input_text: The text to generate OpenAI embeddings for. This can be a string, a list of strings,
                    a list of integers, or a list of lists of integers.
    :param model: The OpenAI model to be used for generating the embeddings.
    :param embedding_kwargs: Additional keyword arguments to pass to the OpenAI `create_embeddings` method.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:OpenAIEmbeddingOperator`
        For possible options for `embedding_kwargs`, see:
        https://platform.openai.com/docs/api-reference/embeddings/create
    """

    template_fields: Sequence[str] = ("input_text",)

    def __init__(
        self,
        conn_id: str,
        input_text: str | list[str] | list[int] | list[list[int]],
        model: str = "text-embedding-ada-002",
        embedding_kwargs: dict | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.conn_id = conn_id
        self.input_text = input_text
        self.model = model
        self.embedding_kwargs = embedding_kwargs or {}

    @cached_property
    def hook(self) -> OpenAIHook:
        """Return an instance of the OpenAIHook."""
        return OpenAIHook(conn_id=self.conn_id)

    def execute(self, context: Context) -> list[float]:
        if not self.input_text or not isinstance(self.input_text, (str, list)):
            raise ValueError(
                "The 'input_text' must be a non-empty string, list of strings, list of integers, or list of lists of integers."
            )
        self.log.info("Generating embeddings for the input text of length: %d", len(self.input_text))
        embeddings = self.hook.create_embeddings(self.input_text, model=self.model, **self.embedding_kwargs)
        self.log.info("Generated embeddings for %d items", len(embeddings))
        return embeddings


class OpenAITriggerBatchOperator(BaseOperator):
    """
    Operator that triggers an OpenAI Batch API endpoint and waits for the batch to complete.

    :param file_id: Required. The ID of the batch file to trigger.
    :param endpoint: Required. The OpenAI Batch API endpoint to trigger.
    :param conn_id: Optional. The OpenAI connection ID to use. Defaults to 'openai_default'.
    :param deferrable: Optional. Run operator in the deferrable mode.
    :param wait_seconds: Optional. Number of seconds between checks. Only used when ``deferrable`` is False.
        Defaults to 3 seconds.
    :param timeout: Optional. The amount of time, in seconds, to wait for the request to complete.
        Only used when ``deferrable`` is False. Defaults to 24 hour, which is the SLA for OpenAI Batch API.
    :param wait_for_completion: Optional. Whether to wait for the batch to complete. If set to False, the operator
        will return immediately after triggering the batch. Defaults to True.

    .. seealso::
        For more information on how to use this operator, please take a look at the guide:
        :ref:`howto/operator:OpenAITriggerBatchOperator`
    """

    template_fields: Sequence[str] = ("file_id",)

    def __init__(
        self,
        file_id: str,
        endpoint: Literal["/v1/chat/completions", "/v1/embeddings", "/v1/completions"],
        conn_id: str = OpenAIHook.default_conn_name,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        wait_seconds: float = 3,
        timeout: float = 24 * 60 * 60,
        wait_for_completion: bool = True,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.conn_id = conn_id
        self.file_id = file_id
        self.endpoint = endpoint
        self.deferrable = deferrable
        self.wait_seconds = wait_seconds
        self.timeout = timeout
        self.wait_for_completion = wait_for_completion
        self.batch_id: str | None = None

    @cached_property
    def hook(self) -> OpenAIHook:
        """Return an instance of the OpenAIHook."""
        return OpenAIHook(conn_id=self.conn_id)

    def execute(self, context: Context) -> str | None:
        batch = self.hook.create_batch(file_id=self.file_id, endpoint=self.endpoint)
        self.batch_id = batch.id
        if self.wait_for_completion:
            if self.deferrable:
                self.defer(
                    timeout=self.execution_timeout,
                    trigger=OpenAIBatchTrigger(
                        conn_id=self.conn_id,
                        batch_id=self.batch_id,
                        poll_interval=60,
                        end_time=time.time() + self.timeout,
                    ),
                    method_name="execute_complete",
                )
            else:
                self.log.info("Waiting for batch %s to complete", self.batch_id)
                self.hook.wait_for_batch(self.batch_id, wait_seconds=self.wait_seconds, timeout=self.timeout)
        return self.batch_id

    def execute_complete(self, context: Context, event: Any = None) -> str:
        """
        Invoke this callback when the trigger fires; return immediately.

        Relies on trigger to throw an exception, otherwise it assumes execution was
        successful.
        """
        if event["status"] == "error":
            raise OpenAIBatchJobException(event["message"])

        self.log.info("%s completed successfully.", self.task_id)
        return event["batch_id"]

    def on_kill(self) -> None:
        """Cancel the batch if task is cancelled."""
        if self.batch_id:
            self.log.info("on_kill: cancel the OpenAI Batch %s", self.batch_id)
            self.hook.cancel_batch(self.batch_id)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any, Literal

try:
    from airflow.sdk import dag, task
except ImportError:
    # Airflow 2 path
    from airflow.decorators import dag, task  # type: ignore[attr-defined,no-redef]

OPENAI_CONN_ID = "openai_default"

POKEMONS = [
    "pikachu",
    "charmander",
    "bulbasaur",
]


@dag(
    schedule=None,
    catchup=False,
)
def openai_batch_chat_completions():
    @task
    def generate_messages(pokemon, **context) -> list[dict[str, Any]]:
        return [{"role": "user", "content": f"Describe the info about {pokemon}?"}]

    @task
    def batch_upload(messages_batch, **context) -> str:
        import tempfile
        import uuid

        from pydantic import BaseModel, Field

        from airflow.providers.openai.hooks.openai import OpenAIHook

        class RequestBody(BaseModel):
            model: str
            messages: list[dict[str, Any]]
            max_tokens: int = Field(default=1000)

        class BatchModel(BaseModel):
            custom_id: str
            method: Literal["POST"]
            url: Literal["/v1/chat/completions"]
            body: RequestBody

        model = "gpt-4o-mini"
        max_tokens = 1000
        hook = OpenAIHook(conn_id=OPENAI_CONN_ID)
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as file:
            for messages in messages_batch:
                file.write(
                    BatchModel(
                        custom_id=str(uuid.uuid4()),
                        method="POST",
                        url="/v1/chat/completions",
                        body=RequestBody(
                            model=model,
                            max_tokens=max_tokens,
                            messages=messages,
                        ),
                    ).model_dump_json()
                    + "\n"
                )
        batch_file = hook.upload_file(file.name, purpose="batch")
        return batch_file.id

    @task
    def cleanup_batch_output_file(batch_id, **context):
        from airflow.providers.openai.hooks.openai import OpenAIHook

        hook = OpenAIHook(conn_id=OPENAI_CONN_ID)
        batch = hook.get_batch(batch_id)
        if batch.output_file_id:
            hook.delete_file(batch.output_file_id)

    messages = generate_messages.expand(pokemon=POKEMONS)
    batch_file_id = batch_upload(messages_batch=messages)

    # [START howto_operator_openai_trigger_operator]
    from airflow.providers.openai.operators.openai import OpenAITriggerBatchOperator

    batch_id = OpenAITriggerBatchOperator(
        task_id="batch_operator_deferred",
        conn_id=OPENAI_CONN_ID,
        file_id=batch_file_id,
        endpoint="/v1/chat/completions",
        deferrable=True,
    )
    # [END howto_operator_openai_trigger_operator]
    cleanup_batch_output = cleanup_batch_output_file(
        batch_id="{{ ti.xcom_pull(task_ids='batch_operator_deferred', key='return_value') }}"
    )
    batch_id >> cleanup_batch_output


openai_batch_chat_completions()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest.mock import Mock

import pytest

from airflow.providers.openai.exceptions import OpenAIBatchJobException, OpenAIBatchTimeout


@pytest.mark.parametrize(
    "exception_class",
    [
        OpenAIBatchTimeout,
        OpenAIBatchJobException,
    ],
)
def test_wait_for_batch_raise_exception(exception_class):
    mock_hook_instance = Mock()
    mock_hook_instance.wait_for_batch.side_effect = exception_class
    hook = mock_hook_instance
    with pytest.raises(exception_class):
        hook.wait_for_batch(batch_id="batch_id")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from unittest.mock import mock_open, patch

import pytest
from openai.pagination import SyncCursorPage
from openai.types import (
    Batch,
    CreateEmbeddingResponse,
    Embedding,
    FileDeleted,
    FileObject,
    VectorStore,
    VectorStoreDeleted,
)
from openai.types.beta import Assistant, AssistantDeleted, Thread, ThreadDeleted
from openai.types.beta.threads import Message, Run
from openai.types.chat import ChatCompletion
from openai.types.vector_stores import VectorStoreFile, VectorStoreFileBatch, VectorStoreFileDeleted

from airflow.models import Connection
from airflow.providers.openai.exceptions import OpenAIBatchJobException, OpenAIBatchTimeout
from airflow.providers.openai.hooks.openai import OpenAIHook

ASSISTANT_ID = "test_assistant_abc123"
ASSISTANT_NAME = "Test Assistant"
ASSISTANT_INSTRUCTIONS = "You are a test assistant."
THREAD_ID = "test_thread_abc123"
MESSAGE_ID = "test_message_abc123"
RUN_ID = "test_run_abc123"
MODEL = "gpt-4"
FILE_ID = "test_file_abc123"
FILE_NAME = "test_file.pdf"
METADATA = {"modified": "true", "user": "abc123"}
VECTOR_STORE_ID = "test_vs_abc123"
VECTOR_STORE_NAME = "Test Vector Store"
VECTOR_FILE_STORE_BATCH_ID = "test_vfsb_abc123"
BATCH_ID = "test_batch_abc123"

openai = pytest.importorskip("openai")


def create_batch(status) -> Batch:
    return Batch(
        id=BATCH_ID,
        object="batch",
        completion_window="24h",
        created_at=1699061776,
        endpoint="/v1/chat/completions",
        input_file_id=FILE_ID,
        status=status,
    )


@pytest.fixture
def mock_openai_connection():
    conn_id = "openai_conn"
    conn = Connection(
        conn_id=conn_id,
        conn_type="openai",
    )
    os.environ[f"AIRFLOW_CONN_{conn.conn_id.upper()}"] = conn.get_uri()
    return conn


@pytest.fixture
def mock_openai_hook(mock_openai_connection):
    with patch("airflow.providers.openai.hooks.openai.OpenAI"):
        yield OpenAIHook(conn_id=mock_openai_connection.conn_id)


@pytest.fixture
def mock_embeddings_response():
    return CreateEmbeddingResponse(
        data=[Embedding(embedding=[0.1, 0.2, 0.3], index=0, object="embedding")],
        model="text-embedding-ada-002-v2",
        object="list",
        usage={"prompt_tokens": 4, "total_tokens": 4},
    )


@pytest.fixture
def mock_completion():
    return ChatCompletion(
        id="chatcmpl-123",
        object="chat.completion",
        created=1677652288,
        model=MODEL,
        choices=[
            {
                "index": 0,
                "message": {
                    "role": "assistant",
                    "content": "Hello there, how may I assist you today?",
                },
                "logprobs": None,
                "finish_reason": "stop",
            }
        ],
    )


@pytest.fixture
def mock_assistant():
    return Assistant(
        id=ASSISTANT_ID,
        name=ASSISTANT_NAME,
        object="assistant",
        created_at=1677652288,
        model=MODEL,
        instructions=ASSISTANT_INSTRUCTIONS,
        tools=[],
        file_ids=[],
        metadata={},
    )


@pytest.fixture
def mock_assistant_list(mock_assistant):
    return SyncCursorPage[Assistant](data=[mock_assistant])


@pytest.fixture
def mock_thread():
    return Thread(id=THREAD_ID, object="thread", created_at=1698984975, metadata={})


@pytest.fixture
def mock_message():
    return Message(
        id=MESSAGE_ID,
        object="thread.message",
        created_at=1698984975,
        thread_id=THREAD_ID,
        status="completed",
        role="user",
        content=[{"type": "text", "text": {"value": "Tell me something interesting.", "annotations": []}}],
        assistant_id=ASSISTANT_ID,
        run_id=RUN_ID,
        file_ids=[],
        metadata={},
    )


@pytest.fixture
def mock_message_list(mock_message):
    return SyncCursorPage[Message](data=[mock_message])


@pytest.fixture
def mock_run():
    return Run(
        id=RUN_ID,
        object="thread.run",
        created_at=1698107661,
        assistant_id=ASSISTANT_ID,
        parallel_tool_calls=False,
        thread_id=THREAD_ID,
        status="completed",
        started_at=1699073476,
        completed_at=1699073476,
        model=MODEL,
        instructions="You are a test assistant.",
        tools=[],
        file_ids=[],
        metadata={},
    )


@pytest.fixture
def mock_run_list(mock_run):
    return SyncCursorPage[Run](data=[mock_run])


@pytest.fixture
def mock_file():
    return FileObject(
        id=FILE_ID,
        object="file",
        bytes=120000,
        created_at=1677610602,
        filename=FILE_NAME,
        purpose="assistants",
        status="processed",
    )


@pytest.fixture
def mock_file_list(mock_file):
    return SyncCursorPage[FileObject](data=[mock_file])


@pytest.fixture
def mock_vector_store():
    return VectorStore(
        id=VECTOR_STORE_ID,
        object="vector_store",
        created_at=1698107661,
        usage_bytes=123456,
        last_active_at=1698107661,
        name=VECTOR_STORE_NAME,
        bytes=123456,
        status="completed",
        file_counts={"in_progress": 0, "completed": 100, "cancelled": 0, "failed": 0, "total": 100},
        metadata={},
        last_used_at=1698107661,
    )


@pytest.fixture
def mock_vector_store_list(mock_vector_store):
    return SyncCursorPage[VectorStore](data=[mock_vector_store])


@pytest.fixture
def mock_vector_file_store_batch():
    return VectorStoreFileBatch(
        id=VECTOR_FILE_STORE_BATCH_ID,
        object="vector_store.files_batch",
        created_at=1699061776,
        vector_store_id=VECTOR_STORE_ID,
        status="completed",
        file_counts={
            "in_progress": 0,
            "completed": 3,
            "failed": 0,
            "cancelled": 0,
            "total": 0,
        },
    )


@pytest.fixture
def mock_vector_file_store_list():
    return SyncCursorPage[VectorStoreFile](
        data=[
            VectorStoreFile(
                id="test-file-abc123",
                object="vector_store.file",
                created_at=1699061776,
                usage_bytes=1234,
                vector_store_id=VECTOR_STORE_ID,
                status="completed",
                last_error=None,
            ),
            VectorStoreFile(
                id="test-file-abc456",
                object="vector_store.file",
                created_at=1699061776,
                usage_bytes=1234,
                vector_store_id=VECTOR_STORE_ID,
                status="completed",
                last_error=None,
            ),
        ]
    )


@pytest.fixture(
    params=[
        "completed",
        "expired",
        "cancelling",
        "cancelled",
        "failed",
    ]
)
def mock_terminated_batch(request):
    return create_batch(request.param)


@pytest.fixture(params=["validating", "in_progress", "finalizing"])
def mock_wip_batch(request):
    return create_batch(request.param)


def test_create_chat_completion(mock_openai_hook, mock_completion):
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Hello!"},
    ]

    mock_openai_hook.conn.chat.completions.create.return_value = mock_completion
    completion = mock_openai_hook.create_chat_completion(model=MODEL, messages=messages)
    choice = completion[0]
    assert choice.message.content == "Hello there, how may I assist you today?"


def test_create_assistant(mock_openai_hook, mock_assistant):
    mock_openai_hook.conn.beta.assistants.create.return_value = mock_assistant
    assistant = mock_openai_hook.create_assistant(
        name=ASSISTANT_NAME, model=MODEL, instructions=ASSISTANT_INSTRUCTIONS
    )
    assert assistant.name == ASSISTANT_NAME
    assert assistant.model == MODEL
    assert assistant.instructions == ASSISTANT_INSTRUCTIONS


def test_get_assistant(mock_openai_hook, mock_assistant):
    mock_openai_hook.conn.beta.assistants.retrieve.return_value = mock_assistant
    assistant = mock_openai_hook.get_assistant(assistant_id=ASSISTANT_ID)
    assert assistant.name == ASSISTANT_NAME
    assert assistant.model == MODEL
    assert assistant.instructions == ASSISTANT_INSTRUCTIONS


def test_get_assistants(mock_openai_hook, mock_assistant_list):
    mock_openai_hook.conn.beta.assistants.list.return_value = mock_assistant_list
    assistants = mock_openai_hook.get_assistants()
    assert isinstance(assistants, list)


def test_modify_assistant(mock_openai_hook, mock_assistant):
    new_assistant_name = "New Test Assistant"
    mock_assistant.name = new_assistant_name
    mock_openai_hook.conn.beta.assistants.update.return_value = mock_assistant
    assistant = mock_openai_hook.modify_assistant(assistant_id=ASSISTANT_ID, name=new_assistant_name)
    assert assistant.name == new_assistant_name


def test_delete_assistant(mock_openai_hook):
    delete_response = AssistantDeleted(id=ASSISTANT_ID, object="assistant.deleted", deleted=True)
    mock_openai_hook.conn.beta.assistants.delete.return_value = delete_response
    assistant_deleted = mock_openai_hook.delete_assistant(assistant_id=ASSISTANT_ID)
    assert assistant_deleted.deleted


def test_create_thread(mock_openai_hook, mock_thread):
    mock_openai_hook.conn.beta.threads.create.return_value = mock_thread
    thread = mock_openai_hook.create_thread()
    assert thread.id == THREAD_ID


def test_modify_thread(mock_openai_hook, mock_thread):
    mock_thread.metadata = METADATA
    mock_openai_hook.conn.beta.threads.update.return_value = mock_thread
    thread = mock_openai_hook.modify_thread(thread_id=THREAD_ID, metadata=METADATA)
    assert thread.metadata.get("modified") == "true"
    assert thread.metadata.get("user") == "abc123"


def test_delete_thread(mock_openai_hook):
    delete_response = ThreadDeleted(id=THREAD_ID, object="thread.deleted", deleted=True)
    mock_openai_hook.conn.beta.threads.delete.return_value = delete_response
    thread_deleted = mock_openai_hook.delete_thread(thread_id=THREAD_ID)
    assert thread_deleted.deleted


def test_create_message(mock_openai_hook, mock_message):
    role = "user"
    content = "Tell me something interesting."
    mock_openai_hook.conn.beta.threads.messages.create.return_value = mock_message
    message = mock_openai_hook.create_message(thread_id=THREAD_ID, content=content, role=role)
    assert message.id == MESSAGE_ID


def test_get_messages(mock_openai_hook, mock_message_list):
    mock_openai_hook.conn.beta.threads.messages.list.return_value = mock_message_list
    messages = mock_openai_hook.get_messages(thread_id=THREAD_ID)
    assert isinstance(messages, list)


def test_modify_messages(mock_openai_hook, mock_message):
    mock_message.metadata = METADATA
    mock_openai_hook.conn.beta.threads.messages.update.return_value = mock_message
    message = mock_openai_hook.modify_message(thread_id=THREAD_ID, message_id=MESSAGE_ID, metadata=METADATA)
    assert message.metadata.get("modified") == "true"
    assert message.metadata.get("user") == "abc123"


def test_create_run(mock_openai_hook, mock_run):
    thread_id = THREAD_ID
    assistant_id = ASSISTANT_ID
    mock_openai_hook.conn.beta.threads.runs.create.return_value = mock_run
    run = mock_openai_hook.create_run(thread_id=thread_id, assistant_id=assistant_id)
    assert run.id == RUN_ID


def test_create_run_and_poll(mock_openai_hook, mock_run):
    thread_id = THREAD_ID
    assistant_id = ASSISTANT_ID
    mock_openai_hook.conn.beta.threads.runs.create_and_poll.return_value = mock_run
    run = mock_openai_hook.create_run_and_poll(thread_id=thread_id, assistant_id=assistant_id)
    assert run.id == RUN_ID


def test_get_runs(mock_openai_hook, mock_run_list):
    mock_openai_hook.conn.beta.threads.runs.list.return_value = mock_run_list
    runs = mock_openai_hook.get_runs(thread_id=THREAD_ID)
    assert isinstance(runs, list)


def test_get_run_with_run_id(mock_openai_hook, mock_run):
    mock_openai_hook.conn.beta.threads.runs.retrieve.return_value = mock_run
    run = mock_openai_hook.get_run(thread_id=THREAD_ID, run_id=RUN_ID)
    assert run.id == RUN_ID


def test_modify_run(mock_openai_hook, mock_run):
    mock_run.metadata = METADATA
    mock_openai_hook.conn.beta.threads.runs.update.return_value = mock_run
    message = mock_openai_hook.modify_run(thread_id=THREAD_ID, run_id=RUN_ID, metadata=METADATA)
    assert message.metadata.get("modified") == "true"
    assert message.metadata.get("user") == "abc123"


def test_create_embeddings(mock_openai_hook, mock_embeddings_response):
    text = "Sample text"
    mock_openai_hook.conn.embeddings.create.return_value = mock_embeddings_response
    embeddings = mock_openai_hook.create_embeddings(text)
    assert embeddings == [0.1, 0.2, 0.3]


@patch("builtins.open", new_callable=mock_open, read_data="test-data")
def test_upload_file(mock_file_open, mock_openai_hook, mock_file):
    mock_file.name = FILE_NAME
    mock_file.purpose = "assistants"
    mock_openai_hook.conn.files.create.return_value = mock_file
    file = mock_openai_hook.upload_file(file=mock_file_open(), purpose="assistants")
    assert file.name == FILE_NAME
    assert file.purpose == "assistants"


def test_get_file(mock_openai_hook, mock_file):
    mock_openai_hook.conn.files.retrieve.return_value = mock_file
    file = mock_openai_hook.get_file(file_id=FILE_ID)
    assert file.id == FILE_ID
    assert file.filename == FILE_NAME


def test_get_files(mock_openai_hook, mock_file_list):
    mock_openai_hook.conn.files.list.return_value = mock_file_list
    files = mock_openai_hook.get_files()
    assert isinstance(files, list)


def test_delete_file(mock_openai_hook):
    delete_response = FileDeleted(id=FILE_ID, object="file", deleted=True)
    mock_openai_hook.conn.files.delete.return_value = delete_response
    file_deleted = mock_openai_hook.delete_file(file_id=FILE_ID)
    assert file_deleted.deleted


def test_create_vector_store(mock_openai_hook, mock_vector_store):
    mock_openai_hook.conn.vector_stores.create.return_value = mock_vector_store
    vector_store = mock_openai_hook.create_vector_store(name=VECTOR_STORE_NAME)
    assert vector_store.id == VECTOR_STORE_ID
    assert vector_store.name == VECTOR_STORE_NAME


def test_get_vector_store(mock_openai_hook, mock_vector_store):
    mock_openai_hook.conn.vector_stores.retrieve.return_value = mock_vector_store
    vector_store = mock_openai_hook.get_vector_store(vector_store_id=VECTOR_STORE_ID)
    assert vector_store.id == VECTOR_STORE_ID
    assert vector_store.name == VECTOR_STORE_NAME


def test_get_vector_stores(mock_openai_hook, mock_vector_store_list):
    mock_openai_hook.conn.vector_stores.list.return_value = mock_vector_store_list
    vector_stores = mock_openai_hook.get_vector_stores()
    assert isinstance(vector_stores, list)


def test_modify_vector_store(mock_openai_hook, mock_vector_store):
    new_vector_store_name = "New Vector Store"
    mock_vector_store.name = new_vector_store_name
    mock_openai_hook.conn.vector_stores.update.return_value = mock_vector_store
    vector_store = mock_openai_hook.modify_vector_store(
        vector_store_id=VECTOR_STORE_ID, name=new_vector_store_name
    )
    assert vector_store.name == new_vector_store_name


def test_delete_vector_store(mock_openai_hook):
    delete_response = VectorStoreDeleted(id=VECTOR_STORE_ID, object="vector_store.deleted", deleted=True)
    mock_openai_hook.conn.vector_stores.delete.return_value = delete_response
    vector_store_deleted = mock_openai_hook.delete_vector_store(vector_store_id=VECTOR_STORE_ID)
    assert vector_store_deleted.deleted


def test_upload_files_to_vector_store(mock_openai_hook, mock_vector_file_store_batch):
    files = ["file1.txt", "file2.txt", "file3.txt"]
    mock_openai_hook.conn.vector_stores.file_batches.upload_and_poll.return_value = (
        mock_vector_file_store_batch
    )
    vector_file_store_batch = mock_openai_hook.upload_files_to_vector_store(
        vector_store_id=VECTOR_STORE_ID, files=files
    )
    assert vector_file_store_batch.id == VECTOR_FILE_STORE_BATCH_ID
    assert vector_file_store_batch.file_counts.completed == len(files)


def test_get_vector_store_files(mock_openai_hook, mock_vector_file_store_list):
    mock_openai_hook.conn.vector_stores.files.list.return_value = mock_vector_file_store_list
    vector_file_store_list = mock_openai_hook.get_vector_store_files(vector_store_id=VECTOR_STORE_ID)
    assert isinstance(vector_file_store_list, list)


def test_delete_vector_store_file(mock_openai_hook):
    delete_response = VectorStoreFileDeleted(
        id="test_file_abc123", object="vector_store.file.deleted", deleted=True
    )
    mock_openai_hook.conn.vector_stores.files.delete.return_value = delete_response
    vector_store_file_deleted = mock_openai_hook.delete_vector_store_file(
        vector_store_id=VECTOR_STORE_ID, file_id=FILE_ID
    )
    assert vector_store_file_deleted.id == FILE_ID
    assert vector_store_file_deleted.deleted


def test_create_batch(mock_openai_hook, mock_terminated_batch):
    mock_openai_hook.conn.batches.create.return_value = mock_terminated_batch
    batch = mock_openai_hook.create_batch(endpoint="/v1/chat/completions", file_id=FILE_ID)
    assert batch.id == mock_terminated_batch.id


def test_get_batch(mock_openai_hook, mock_terminated_batch):
    mock_openai_hook.conn.batches.retrieve.return_value = mock_terminated_batch
    batch = mock_openai_hook.get_batch(batch_id=BATCH_ID)
    assert batch.id == mock_terminated_batch.id


def test_cancel_batch(mock_openai_hook, mock_terminated_batch):
    mock_openai_hook.conn.batches.cancel.return_value = mock_terminated_batch
    batch = mock_openai_hook.cancel_batch(batch_id=BATCH_ID)
    assert batch.id == mock_terminated_batch.id


def test_wait_for_finished_batch(mock_openai_hook, mock_terminated_batch):
    mock_openai_hook.conn.batches.retrieve.return_value = mock_terminated_batch
    if mock_terminated_batch.status == "completed":
        try:
            mock_openai_hook.wait_for_batch(batch_id=BATCH_ID)
        except Exception as e:
            pytest.fail(f"Should not have raised exception: {e}")
    else:
        with pytest.raises(OpenAIBatchJobException, match="Batch failed"):
            mock_openai_hook.wait_for_batch(batch_id=BATCH_ID, wait_seconds=0.01, timeout=0.1)


def test_wait_for_in_progress_batch_timeout(mock_openai_hook, mock_wip_batch):
    mock_openai_hook.conn.batches.retrieve.return_value = mock_wip_batch
    with pytest.raises(OpenAIBatchTimeout, match="Timeout"):
        mock_openai_hook.wait_for_batch(batch_id=BATCH_ID, wait_seconds=0.2, timeout=0.01)
    assert mock_openai_hook.conn.batches.retrieve.call_count >= 1
    assert mock_openai_hook.conn.batches.cancel.call_count == 1


def test_openai_hook_test_connection(mock_openai_hook):
    result, message = mock_openai_hook.test_connection()
    assert result is True
    assert message == "Connection established!"


@patch("airflow.providers.openai.hooks.openai.OpenAI")
def test_get_conn_with_api_key_in_extra(mock_client):
    conn_id = "api_key_in_extra"
    conn = Connection(
        conn_id=conn_id,
        conn_type="openai",
        extra={"openai_client_kwargs": {"api_key": "api_key_in_extra"}},
    )
    os.environ[f"AIRFLOW_CONN_{conn.conn_id.upper()}"] = conn.get_uri()
    hook = OpenAIHook(conn_id=conn_id)
    hook.get_conn()
    mock_client.assert_called_once_with(
        api_key="api_key_in_extra",
        base_url=None,
    )


@patch("airflow.providers.openai.hooks.openai.OpenAI")
def test_get_conn_with_api_key_in_password(mock_client):
    conn_id = "api_key_in_password"
    conn = Connection(
        conn_id=conn_id,
        conn_type="openai",
        password="api_key_in_password",
    )
    os.environ[f"AIRFLOW_CONN_{conn.conn_id.upper()}"] = conn.get_uri()
    hook = OpenAIHook(conn_id=conn_id)
    hook.get_conn()
    mock_client.assert_called_once_with(
        api_key="api_key_in_password",
        base_url=None,
    )


@patch("airflow.providers.openai.hooks.openai.OpenAI")
def test_get_conn_with_base_url_in_extra(mock_client):
    conn_id = "base_url_in_extra"
    conn = Connection(
        conn_id=conn_id,
        conn_type="openai",
        extra={"openai_client_kwargs": {"base_url": "base_url_in_extra", "api_key": "api_key_in_extra"}},
    )
    os.environ[f"AIRFLOW_CONN_{conn.conn_id.upper()}"] = conn.get_uri()
    hook = OpenAIHook(conn_id=conn_id)
    hook.get_conn()
    mock_client.assert_called_once_with(
        api_key="api_key_in_extra",
        base_url="base_url_in_extra",
    )


@patch("airflow.providers.openai.hooks.openai.OpenAI")
def test_get_conn_with_openai_client_kwargs(mock_client):
    conn_id = "openai_client_kwargs"
    conn = Connection(
        conn_id=conn_id,
        conn_type="openai",
        extra={
            "openai_client_kwargs": {
                "api_key": "api_key_in_extra",
                "organization": "organization_in_extra",
            }
        },
    )
    os.environ[f"AIRFLOW_CONN_{conn.conn_id.upper()}"] = conn.get_uri()
    hook = OpenAIHook(conn_id=conn_id)
    hook.get_conn()
    mock_client.assert_called_once_with(
        api_key="api_key_in_extra",
        base_url=None,
        organization="organization_in_extra",
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import Mock

import pytest
from openai.types.batch import Batch

from airflow.exceptions import TaskDeferred
from airflow.providers.common.compat.sdk import Context
from airflow.providers.openai.operators.openai import OpenAIEmbeddingOperator, OpenAITriggerBatchOperator
from airflow.providers.openai.triggers.openai import OpenAIBatchTrigger

openai = pytest.importorskip("openai")
TASK_ID = "TaskId"
CONN_ID = "test_conn_id"
BATCH_ID = "batch_id"
FILE_ID = "file_id"
BATCH_ENDPOINT = "/v1/chat/completions"


@pytest.fixture
def mock_batch():
    return Batch(
        id=BATCH_ID,
        object="batch",
        completion_window="24h",
        created_at=1699061776,
        endpoint=BATCH_ENDPOINT,
        input_file_id=FILE_ID,
        status="in_progress",
    )


def test_execute_with_input_text():
    operator = OpenAIEmbeddingOperator(
        task_id=TASK_ID, conn_id=CONN_ID, model="test_model", input_text="Test input text"
    )
    mock_hook_instance = Mock()
    mock_hook_instance.create_embeddings.return_value = [1.0, 2.0, 3.0]
    operator.hook = mock_hook_instance

    context = Context()
    embeddings = operator.execute(context)

    assert embeddings == [1.0, 2.0, 3.0]


@pytest.mark.parametrize("invalid_input", ["", None, 123])
def test_execute_with_invalid_input(invalid_input):
    operator = OpenAIEmbeddingOperator(
        task_id=TASK_ID, conn_id=CONN_ID, model="test_model", input_text=invalid_input
    )
    context = Context()
    with pytest.raises(
        ValueError,
        match="The 'input_text' must be a non-empty string, list of strings, list of integers, or list of lists of integers.",
    ):
        operator.execute(context)


@pytest.mark.parametrize("wait_for_completion", [True, False])
def test_openai_trigger_batch_operator_not_deferred(mock_batch, wait_for_completion):
    operator = OpenAITriggerBatchOperator(
        task_id=TASK_ID,
        conn_id=CONN_ID,
        file_id=FILE_ID,
        endpoint=BATCH_ENDPOINT,
        wait_for_completion=wait_for_completion,
        deferrable=False,
    )
    mock_hook_instance = Mock()
    mock_hook_instance.get_batch.return_value = mock_batch
    mock_hook_instance.create_batch.return_value = mock_batch
    operator.hook = mock_hook_instance

    context = Context()
    batch_id = operator.execute(context)
    assert batch_id == BATCH_ID


@pytest.mark.parametrize("wait_for_completion", [True, False])
def test_openai_trigger_batch_operator_with_deferred(mock_batch, wait_for_completion):
    operator = OpenAITriggerBatchOperator(
        task_id=TASK_ID,
        conn_id=CONN_ID,
        file_id=FILE_ID,
        endpoint=BATCH_ENDPOINT,
        deferrable=True,
        wait_for_completion=wait_for_completion,
    )
    mock_hook_instance = Mock()
    mock_hook_instance.get_batch.return_value = mock_batch
    mock_hook_instance.create_batch.return_value = mock_batch
    operator.hook = mock_hook_instance

    context = Context()
    if wait_for_completion:
        with pytest.raises(TaskDeferred) as exc:
            operator.execute(context)
        assert isinstance(exc.value.trigger, OpenAIBatchTrigger)
    else:
        batch_id = operator.execute(context)
        assert batch_id == BATCH_ID



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import time
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Any, BinaryIO, Literal

from openai import OpenAI

if TYPE_CHECKING:
    from openai.types import (
        FileDeleted,
        FileObject,
        VectorStore,
        VectorStoreDeleted,
    )
    from openai.types.batch import Batch
    from openai.types.beta import Assistant, AssistantDeleted, Thread, ThreadDeleted
    from openai.types.beta.threads import Message, Run
    from openai.types.chat import (
        ChatCompletionAssistantMessageParam,
        ChatCompletionFunctionMessageParam,
        ChatCompletionMessage,
        ChatCompletionSystemMessageParam,
        ChatCompletionToolMessageParam,
        ChatCompletionUserMessageParam,
    )
    from openai.types.vector_stores import VectorStoreFile, VectorStoreFileBatch, VectorStoreFileDeleted
from airflow.providers.common.compat.sdk import BaseHook
from airflow.providers.openai.exceptions import OpenAIBatchJobException, OpenAIBatchTimeout


class BatchStatus(str, Enum):
    """Enum for the status of a batch."""

    VALIDATING = "validating"
    FAILED = "failed"
    IN_PROGRESS = "in_progress"
    FINALIZING = "finalizing"
    COMPLETED = "completed"
    EXPIRED = "expired"
    CANCELLING = "cancelling"
    CANCELLED = "cancelled"

    def __str__(self) -> str:
        return str(self.value)

    @classmethod
    def is_in_progress(cls, status: str) -> bool:
        """Check if the batch status is in progress."""
        return status in (cls.VALIDATING, cls.IN_PROGRESS, cls.FINALIZING)


class OpenAIHook(BaseHook):
    """
    Use OpenAI SDK to interact with OpenAI APIs.

    .. seealso:: https://platform.openai.com/docs/introduction/overview

    :param conn_id: :ref:`OpenAI connection id <howto/connection:openai>`
    """

    conn_name_attr = "conn_id"
    default_conn_name = "openai_default"
    conn_type = "openai"
    hook_name = "OpenAI"

    def __init__(self, conn_id: str = default_conn_name, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.conn_id = conn_id

    @classmethod
    def get_ui_field_behaviour(cls) -> dict[str, Any]:
        """Return custom field behaviour."""
        return {
            "hidden_fields": ["schema", "port", "login"],
            "relabeling": {"password": "API Key"},
            "placeholders": {},
        }

    def test_connection(self) -> tuple[bool, str]:
        try:
            self.conn.models.list()
            return True, "Connection established!"
        except Exception as e:
            return False, str(e)

    @cached_property
    def conn(self) -> OpenAI:
        """Return an OpenAI connection object."""
        return self.get_conn()

    def get_conn(self) -> OpenAI:
        """Return an OpenAI connection object."""
        conn = self.get_connection(self.conn_id)
        extras = conn.extra_dejson
        openai_client_kwargs = extras.get("openai_client_kwargs", {})
        api_key = openai_client_kwargs.pop("api_key", None) or conn.password
        base_url = openai_client_kwargs.pop("base_url", None) or conn.host or None
        return OpenAI(
            api_key=api_key,
            base_url=base_url,
            **openai_client_kwargs,
        )

    def create_chat_completion(
        self,
        messages: list[
            ChatCompletionSystemMessageParam
            | ChatCompletionUserMessageParam
            | ChatCompletionAssistantMessageParam
            | ChatCompletionToolMessageParam
            | ChatCompletionFunctionMessageParam
        ],
        model: str = "gpt-3.5-turbo",
        **kwargs: Any,
    ) -> list[ChatCompletionMessage]:
        """
        Create a model response for the given chat conversation and returns a list of chat completions.

        :param messages: A list of messages comprising the conversation so far
        :param model: ID of the model to use
        """
        response = self.conn.chat.completions.create(model=model, messages=messages, **kwargs)
        return response.choices

    def create_assistant(self, model: str = "gpt-3.5-turbo", **kwargs: Any) -> Assistant:
        """
        Create an OpenAI assistant using the given model.

        :param model: The OpenAI model for the assistant to use.
        """
        assistant = self.conn.beta.assistants.create(model=model, **kwargs)
        return assistant

    def get_assistant(self, assistant_id: str) -> Assistant:
        """
        Get an OpenAI assistant.

        :param assistant_id: The ID of the assistant to retrieve.
        """
        assistant = self.conn.beta.assistants.retrieve(assistant_id=assistant_id)
        return assistant

    def get_assistants(self, **kwargs: Any) -> list[Assistant]:
        """Get a list of Assistant objects."""
        assistants = self.conn.beta.assistants.list(**kwargs)
        return assistants.data

    def modify_assistant(self, assistant_id: str, **kwargs: Any) -> Assistant:
        """
        Modify an existing Assistant object.

        :param assistant_id: The ID of the assistant to be modified.
        """
        assistant = self.conn.beta.assistants.update(assistant_id=assistant_id, **kwargs)
        return assistant

    def delete_assistant(self, assistant_id: str) -> AssistantDeleted:
        """
        Delete an OpenAI Assistant for a given ID.

        :param assistant_id: The ID of the assistant to delete.
        """
        response = self.conn.beta.assistants.delete(assistant_id=assistant_id)
        return response

    def create_thread(self, **kwargs: Any) -> Thread:
        """Create an OpenAI thread."""
        thread = self.conn.beta.threads.create(**kwargs)
        return thread

    def modify_thread(self, thread_id: str, metadata: dict[str, Any]) -> Thread:
        """
        Modify an existing Thread object.

        :param thread_id: The ID of the thread to modify. Only the metadata can be modified.
        :param metadata: Set of 16 key-value pairs that can be attached to an object.
        """
        thread = self.conn.beta.threads.update(thread_id=thread_id, metadata=metadata)
        return thread

    def delete_thread(self, thread_id: str) -> ThreadDeleted:
        """
        Delete an OpenAI thread for a given thread_id.

        :param thread_id: The ID of the thread to delete.
        """
        response = self.conn.beta.threads.delete(thread_id=thread_id)
        return response

    def create_message(
        self, thread_id: str, role: Literal["user", "assistant"], content: str, **kwargs: Any
    ) -> Message:
        """
        Create a message for a given Thread.

        :param thread_id: The ID of the thread to create a message for.
        :param role: The role of the entity that is creating the message. Allowed values include: 'user', 'assistant'.
        :param content: The content of the message.
        """
        thread_message = self.conn.beta.threads.messages.create(
            thread_id=thread_id, role=role, content=content, **kwargs
        )
        return thread_message

    def get_messages(self, thread_id: str, **kwargs: Any) -> list[Message]:
        """
        Return a list of messages for a given Thread.

        :param thread_id: The ID of the thread the messages belong to.
        """
        messages = self.conn.beta.threads.messages.list(thread_id=thread_id, **kwargs)
        return messages.data

    def modify_message(self, thread_id: str, message_id, **kwargs: Any) -> Message:
        """
        Modify an existing message for a given Thread.

        :param thread_id: The ID of the thread to which this message belongs.
        :param message_id: The ID of the message to modify.
        """
        thread_message = self.conn.beta.threads.messages.update(
            thread_id=thread_id, message_id=message_id, **kwargs
        )
        return thread_message

    def create_run(self, thread_id: str, assistant_id: str, **kwargs: Any) -> Run:
        """
        Create a run for a given thread and assistant.

        :param thread_id: The ID of the thread to run.
        :param assistant_id: The ID of the assistant to use to execute this run.
        """
        run = self.conn.beta.threads.runs.create(thread_id=thread_id, assistant_id=assistant_id, **kwargs)
        return run

    def create_run_and_poll(self, thread_id: str, assistant_id: str, **kwargs: Any) -> Run:
        """
        Create a run for a given thread and assistant and then polls until completion.

        :param thread_id: The ID of the thread to run.
        :param assistant_id: The ID of the assistant to use to execute this run.
        :return: An OpenAI Run object
        """
        run = self.conn.beta.threads.runs.create_and_poll(
            thread_id=thread_id, assistant_id=assistant_id, **kwargs
        )
        return run

    def get_run(self, thread_id: str, run_id: str) -> Run:
        """
        Retrieve a run for a given thread and run.

        :param thread_id: The ID of the thread that was run.
        :param run_id: The ID of the run to retrieve.
        """
        run = self.conn.beta.threads.runs.retrieve(thread_id=thread_id, run_id=run_id)
        return run

    def get_runs(self, thread_id: str, **kwargs: Any) -> list[Run]:
        """
        Return a list of runs belonging to a thread.

        :param thread_id: The ID of the thread the run belongs to.
        """
        runs = self.conn.beta.threads.runs.list(thread_id=thread_id, **kwargs)
        return runs.data

    def modify_run(self, thread_id: str, run_id: str, **kwargs: Any) -> Run:
        """
        Modify a run on a given thread.

        :param thread_id: The ID of the thread that was run.
        :param run_id: The ID of the run to modify.
        """
        run = self.conn.beta.threads.runs.update(thread_id=thread_id, run_id=run_id, **kwargs)
        return run

    def create_embeddings(
        self,
        text: str | list[str] | list[int] | list[list[int]],
        model: str = "text-embedding-ada-002",
        **kwargs: Any,
    ) -> list[float]:
        """
        Generate embeddings for the given text using the given model.

        :param text: The text to generate embeddings for.
        :param model: The model to use for generating embeddings.
        """
        response = self.conn.embeddings.create(model=model, input=text, **kwargs)
        embeddings: list[float] = response.data[0].embedding
        return embeddings

    def upload_file(self, file: str, purpose: Literal["fine-tune", "assistants", "batch"]) -> FileObject:
        """
        Upload a file that can be used across various endpoints. The size of all the files uploaded by one organization can be up to 100 GB.

        :param file: The File object (not file name) to be uploaded.
        :param purpose: The intended purpose of the uploaded file. Use "fine-tune" for
            Fine-tuning, "assistants" for Assistants and Messages, and "batch" for Batch API.
        """
        with open(file, "rb") as file_stream:
            file_object = self.conn.files.create(file=file_stream, purpose=purpose)
        return file_object

    def get_file(self, file_id: str) -> FileObject:
        """
        Return information about a specific file.

        :param file_id: The ID of the file to use for this request.
        """
        file = self.conn.files.retrieve(file_id=file_id)
        return file

    def get_files(self) -> list[FileObject]:
        """Return a list of files that belong to the user's organization."""
        files = self.conn.files.list()
        return files.data

    def delete_file(self, file_id: str) -> FileDeleted:
        """
        Delete a file.

        :param file_id: The ID of the file to be deleted.
        """
        response = self.conn.files.delete(file_id=file_id)
        return response

    def create_vector_store(self, **kwargs: Any) -> VectorStore:
        """Create a vector store."""
        vector_store = self.conn.vector_stores.create(**kwargs)
        return vector_store

    def get_vector_stores(self, **kwargs: Any) -> list[VectorStore]:
        """Return a list of vector stores."""
        vector_stores = self.conn.vector_stores.list(**kwargs)
        return vector_stores.data

    def get_vector_store(self, vector_store_id: str) -> VectorStore:
        """
        Retrieve a vector store.

        :param vector_store_id: The ID of the vector store to retrieve.
        """
        vector_store = self.conn.vector_stores.retrieve(vector_store_id=vector_store_id)
        return vector_store

    def modify_vector_store(self, vector_store_id: str, **kwargs: Any) -> VectorStore:
        """
        Modify a vector store.

        :param vector_store_id: The ID of the vector store to modify.
        """
        vector_store = self.conn.vector_stores.update(vector_store_id=vector_store_id, **kwargs)
        return vector_store

    def delete_vector_store(self, vector_store_id: str) -> VectorStoreDeleted:
        """
        Delete a vector store.

        :param vector_store_id: The ID of the vector store to delete.
        """
        response = self.conn.vector_stores.delete(vector_store_id=vector_store_id)
        return response

    def upload_files_to_vector_store(
        self, vector_store_id: str, files: list[BinaryIO]
    ) -> VectorStoreFileBatch:
        """
        Upload files to a vector store and poll until completion.

        :param vector_store_id: The ID of the vector store the files are to be uploaded
            to.
        :param files: A list of binary files to upload.
        """
        file_batch = self.conn.vector_stores.file_batches.upload_and_poll(
            vector_store_id=vector_store_id, files=files
        )
        return file_batch

    def get_vector_store_files(self, vector_store_id: str) -> list[VectorStoreFile]:
        """
        Return a list of vector store files.

        :param vector_store_id:
        """
        vector_store_files = self.conn.vector_stores.files.list(vector_store_id=vector_store_id)
        return vector_store_files.data

    def delete_vector_store_file(self, vector_store_id: str, file_id: str) -> VectorStoreFileDeleted:
        """
        Delete a vector store file. This will remove the file from the vector store but the file itself will not be deleted. To delete the file, use delete_file.

        :param vector_store_id: The ID of the vector store that the file belongs to.
        :param file_id: The ID of the file to delete.
        """
        response = self.conn.vector_stores.files.delete(vector_store_id=vector_store_id, file_id=file_id)
        return response

    def create_batch(
        self,
        file_id: str,
        endpoint: Literal["/v1/chat/completions", "/v1/embeddings", "/v1/completions"],
        metadata: dict[str, str] | None = None,
        completion_window: Literal["24h"] = "24h",
    ) -> Batch:
        """
        Create a batch for a given model and files.

        :param file_id: The ID of the file to be used for this batch.
        :param endpoint: The endpoint to use for this batch. Allowed values include:
            '/v1/chat/completions', '/v1/embeddings', '/v1/completions'.
        :param metadata: A set of key-value pairs that can be attached to an object.
        :param completion_window: The time window for the batch to complete. Default is 24 hours.
        """
        batch = self.conn.batches.create(
            input_file_id=file_id, endpoint=endpoint, metadata=metadata, completion_window=completion_window
        )
        return batch

    def get_batch(self, batch_id: str) -> Batch:
        """
        Get the status of a batch.

        :param batch_id: The ID of the batch to get the status of.
        """
        batch = self.conn.batches.retrieve(batch_id=batch_id)
        return batch

    def wait_for_batch(self, batch_id: str, wait_seconds: float = 3, timeout: float = 3600) -> None:
        """
        Poll a batch to check if it finishes.

        :param batch_id: Id of the Batch to wait for.
        :param wait_seconds: Optional. Number of seconds between checks.
        :param timeout: Optional. How many seconds wait for batch to be ready.
            Used only if not ran in deferred operator.
        """
        start = time.monotonic()
        while True:
            if start + timeout < time.monotonic():
                self.cancel_batch(batch_id=batch_id)
                raise OpenAIBatchTimeout(f"Timeout: OpenAI Batch {batch_id} is not ready after {timeout}s")
            batch = self.get_batch(batch_id=batch_id)

            if BatchStatus.is_in_progress(batch.status):
                time.sleep(wait_seconds)
                continue
            if batch.status == BatchStatus.COMPLETED:
                return
            if batch.status == BatchStatus.FAILED:
                raise OpenAIBatchJobException(f"Batch failed - \n{batch_id}")
            if batch.status in (BatchStatus.CANCELLED, BatchStatus.CANCELLING):
                raise OpenAIBatchJobException(f"Batch failed - batch was cancelled:\n{batch_id}")
            if batch.status == BatchStatus.EXPIRED:
                raise OpenAIBatchJobException(
                    f"Batch failed - batch couldn't be completed within the hour time window :\n{batch_id}"
                )

            raise OpenAIBatchJobException(
                f"Batch failed - encountered unexpected status `{batch.status}` for batch_id `{batch_id}`"
            )

    def cancel_batch(self, batch_id: str) -> Batch:
        """
        Cancel a batch.

        :param batch_id: The ID of the batch to delete.
        """
        batch = self.conn.batches.cancel(batch_id=batch_id)
        return batch
